@@ -22,7 +22,9 @@
  *
  */
 
-#include "config.h"
+#include "nm-default.h"
+
+#include "nm-openvpn-service.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -42,23 +44,23 @@
 #include <locale.h>
 #include <pwd.h>
 #include <grp.h>
-#include "glib-unix.h"
+#include <glib-unix.h>
 
-#include "nm-default.h"
-
-#include <nm-vpn-service-plugin.h>
-
-#include "nm-openvpn-service.h"
-#include "nm-utils.h"
 #include "utils.h"
-#include "nm-macros-internal.h"
+#include "nm-utils/nm-shared-utils.h"
+#include "nm-utils/nm-vpn-plugin-macros.h"
 
 #if !defined(DIST_VERSION)
 # define DIST_VERSION VERSION
 #endif
 
+#define RUNDIR  LOCALSTATEDIR"/run/NetworkManager"
+
 static struct {
 	gboolean debug;
+	int log_level;
+	int log_level_ovpn;
+	bool log_syslog;
 	GSList *pids_pending_list;
 } gl/*obal*/;
 
@@ -110,7 +112,7 @@ static ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_CERT,                 G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CIPHER,               G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_KEYSIZE,              G_TYPE_INT, 1, 65535, FALSE },
-	{ NM_OPENVPN_KEY_COMP_LZO,             G_TYPE_BOOLEAN, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_COMP_LZO,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CONNECTION_TYPE,      G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_FLOAT,                G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_FRAGMENT_SIZE,        G_TYPE_INT, 0, G_MAXINT, FALSE },
@@ -120,6 +122,7 @@ static ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_PING,                 G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_PING_EXIT,            G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_PING_RESTART,         G_TYPE_INT, 0, G_MAXINT, FALSE },
+	{ NM_OPENVPN_KEY_MAX_ROUTES,           G_TYPE_INT, 0, 100000000, FALSE },
 	{ NM_OPENVPN_KEY_PROTO_TCP,            G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_PORT,                 G_TYPE_INT, 1, 65535, FALSE },
 	{ NM_OPENVPN_KEY_PROXY_TYPE,           G_TYPE_STRING, 0, 0, FALSE },
@@ -139,7 +142,9 @@ static ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_DEV,                  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_DEV_TYPE,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TUN_IPV6,             G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_TLS_CIPHER,           G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TLS_REMOTE,           G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_VERIFY_X509_NAME,     G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_REMOTE_CERT_TLS,      G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_NS_CERT_TYPE,         G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TUNNEL_MTU,           G_TYPE_INT, 0, G_MAXINT, FALSE },
@@ -161,23 +166,25 @@ static ValidProperty valid_secrets[] = {
 
 /*****************************************************************************/
 
-#define _LOG(log_always, level, ...) \
+#define _NMLOG(level, ...) \
 	G_STMT_START { \
-		if ((log_always) || _LOG_d_enabled ()) { \
-			g_log (G_LOG_DOMAIN, level, __VA_ARGS__); \
+		if (gl.log_level >= (level)) { \
+			g_print ("nm-openvpn[%ld] %-7s " _NM_UTILS_MACRO_FIRST (__VA_ARGS__) "\n", \
+			         (long) getpid (), \
+			         nm_utils_syslog_to_str (level) \
+			         _NM_UTILS_MACRO_REST (__VA_ARGS__)); \
 		} \
 	} G_STMT_END
 
 static gboolean
-_LOG_d_enabled (void)
+_LOGD_enabled (void)
 {
-	return gl.debug;
+	return gl.log_level >= LOG_INFO;
 }
 
-#define _LOG_d(...) _LOG(FALSE, G_LOG_LEVEL_INFO, __VA_ARGS__)
-#define _LOG_D(...) _LOG(TRUE,  G_LOG_LEVEL_INFO, __VA_ARGS__)
-#define _LOG_I(...) _LOG(TRUE,  G_LOG_LEVEL_MESSAGE, __VA_ARGS__)
-#define _LOG_W(...) _LOG(TRUE,  G_LOG_LEVEL_WARNING, __VA_ARGS__)
+#define _LOGD(...) _NMLOG(LOG_INFO,    __VA_ARGS__)
+#define _LOGI(...) _NMLOG(LOG_NOTICE,  __VA_ARGS__)
+#define _LOGW(...) _NMLOG(LOG_WARNING, __VA_ARGS__)
 
 /*****************************************************************************/
 
@@ -216,16 +223,16 @@ pids_pending_child_watch_cb (GPid pid, gint status, gpointer user_data)
 
 		exit_status = WEXITSTATUS (status);
 		if (exit_status != 0)
-			_LOG_W ("openvpn[%ld] exited with error code %d", (long) pid, exit_status);
+			_LOGW ("openvpn[%ld] exited with error code %d", (long) pid, exit_status);
 		else
-			_LOG_I ("openvpn[%ld] exited with success", (long) pid);
+			_LOGI ("openvpn[%ld] exited with success", (long) pid);
 	}
 	else if (WIFSTOPPED (status))
-		_LOG_W ("openvpn[%ld] stopped unexpectedly with signal %d", (long) pid, WSTOPSIG (status));
+		_LOGW ("openvpn[%ld] stopped unexpectedly with signal %d", (long) pid, WSTOPSIG (status));
 	else if (WIFSIGNALED (status))
-		_LOG_W ("openvpn[%ld] died with signal %d", (long) pid, WTERMSIG (status));
+		_LOGW ("openvpn[%ld] died with signal %d", (long) pid, WTERMSIG (status));
 	else
-		_LOG_W ("openvpn[%ld] died from an unnatural cause", (long) pid);
+		_LOGW ("openvpn[%ld] died from an unnatural cause", (long) pid);
 
 	g_return_if_fail (pid_data);
 	g_return_if_fail (pid_data->pid == pid);
@@ -249,7 +256,7 @@ pids_pending_add (GPid pid, NMOpenvpnPlugin *plugin)
 	g_return_if_fail (NM_IS_OPENVPN_PLUGIN (plugin));
 	g_return_if_fail (pid > 0);
 
-	_LOG_I ("openvpn[%ld] started", (long) pid);
+	_LOGI ("openvpn[%ld] started", (long) pid);
 
 	pid_data = g_slice_new (PidsPendingData);
 	pid_data->pid = pid;
@@ -268,7 +275,7 @@ pids_pending_ensure_killed (gpointer user_data)
 
 	g_return_val_if_fail (pid_data && pid_data == pids_pending_get (pid_data->pid), FALSE);
 
-	_LOG_I ("openvpn[%ld]: send SIGKILL", (long) pid_data->pid);
+	_LOGI ("openvpn[%ld]: send SIGKILL", (long) pid_data->pid);
 
 	pid_data->kill_id = 0;
 	kill (pid_data->pid, SIGKILL);
@@ -283,7 +290,7 @@ pids_pending_send_sigterm (GPid pid)
 	pid_data = pids_pending_get (pid);
 	g_return_if_fail (pid_data);
 
-	_LOG_I ("openvpn[%ld]: send SIGTERM", (long) pid);
+	_LOGI ("openvpn[%ld]: send SIGTERM", (long) pid);
 
 	kill (pid, SIGTERM);
 	pid_data->kill_id = g_timeout_add (2000, pids_pending_ensure_killed, pid_data);
@@ -293,7 +300,7 @@ static void
 pids_pending_wait_for_processes (GMainLoop *main_loop)
 {
 	if (gl.pids_pending_list) {
-		_LOG_I ("wait for %u openvpn processes to terminate...", g_slist_length (gl.pids_pending_list));
+		_LOGI ("wait for %u openvpn processes to terminate...", g_slist_length (gl.pids_pending_list));
 
 		do {
 			g_main_context_iteration (g_main_loop_get_context (main_loop), TRUE);
@@ -663,7 +670,7 @@ handle_management_socket (NMOpenvpnPlugin *plugin,
 		return TRUE;
 	}
 
-	_LOG_d ("VPN request '%s'", str);
+	_LOGD ("VPN request '%s'", str);
 
 	auth = get_detail (str, ">PASSWORD:Need '");
 	if (auth) {
@@ -677,13 +684,13 @@ handle_management_socket (NMOpenvpnPlugin *plugin,
 				if (priv->interactive) {
 					gs_free char *joined = NULL;
 
-					_LOG_d ("Requesting new secrets: '%s', %s%s%s", message,
+					_LOGD ("Requesting new secrets: '%s', %s%s%s", message,
 					        NM_PRINT_FMT_QUOTED (hints, "(", (joined = g_strjoinv (",", (char **) hints)), ")", "no hints"));
 
 					nm_vpn_service_plugin_secrets_required ((NMVpnServicePlugin *) plugin, message, (const char **) hints);
 				} else {
 					/* Interactive not allowed, can't ask for more secrets */
-					_LOG_W ("More secrets required but cannot ask interactively");
+					_LOGW ("More secrets required but cannot ask interactively");
 					*out_failure = NM_VPN_PLUGIN_FAILURE_LOGIN_FAILED;
 					again = FALSE;
 				}
@@ -691,7 +698,7 @@ handle_management_socket (NMOpenvpnPlugin *plugin,
 			if (hints)
 				g_free (hints);  /* elements are 'const' */
 		} else {
-			_LOG_W ("Unhandled management socket request '%s'", auth);
+			_LOGW ("Unhandled management socket request '%s'", auth);
 			*out_failure = NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED;
 			again = FALSE;
 		}
@@ -703,7 +710,7 @@ handle_management_socket (NMOpenvpnPlugin *plugin,
 		gboolean fail = TRUE;
 
 		if (!strcmp (auth, "Auth")) {
-			_LOG_W ("Password verification failed");
+			_LOGW ("Password verification failed");
 			if (priv->interactive) {
 				/* Clear existing password in interactive mode, openvpn
 				 * will request a new one after restarting.
@@ -714,9 +721,9 @@ handle_management_socket (NMOpenvpnPlugin *plugin,
 				fail = FALSE;
 			}
 		} else if (!strcmp (auth, "Private Key"))
-			_LOG_W ("Private key verification failed");
+			_LOGW ("Private key verification failed");
 		else
-			_LOG_W ("Unknown verification failed: %s", auth);
+			_LOGW ("Unknown verification failed: %s", auth);
 
 		if (fail) {
 			*out_failure = NM_VPN_PLUGIN_FAILURE_LOGIN_FAILED;
@@ -759,7 +766,7 @@ nm_openvpn_connect_timer_cb (gpointer data)
 	/* open socket and start listener */
 	fd = socket (AF_UNIX, SOCK_STREAM, 0);
 	if (fd < 0) {
-		_LOG_W ("Could not create management socket");
+		_LOGW ("Could not create management socket");
 		nm_vpn_service_plugin_failure (NM_VPN_SERVICE_PLUGIN (plugin), NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
 		goto out;
 	}
@@ -773,7 +780,7 @@ nm_openvpn_connect_timer_cb (gpointer data)
 
 		priv->connect_timer = 0;
 
-		_LOG_W ("Could not open management socket");
+		_LOGW ("Could not open management socket");
 		nm_vpn_service_plugin_failure (NM_VPN_SERVICE_PLUGIN (plugin), NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
 	} else {
 		io_data->socket_channel = g_io_channel_unix_new (fd);
@@ -899,13 +906,6 @@ nm_find_openvpn (void)
 	return *openvpn_binary;
 }
 
-static void
-free_openvpn_args (GPtrArray *args)
-{
-	g_ptr_array_foreach (args, (GFunc) g_free, NULL);
-	g_ptr_array_free (args, TRUE);
-}
-
 static void
 add_openvpn_arg (GPtrArray *args, const char *arg)
 {
@@ -1031,6 +1031,26 @@ update_io_data_from_vpn_setting (NMOpenvpnPluginIOData *io_data,
 	io_data->proxy_password = tmp ? g_strdup (tmp) : NULL;
 }
 
+static char *
+mgt_path_create (NMConnection *connection, GError **error)
+{
+	int errsv;
+
+	/* Setup runtime directory */
+	if (g_mkdir_with_parents (RUNDIR, 0755) != 0) {
+		errsv = errno;
+		g_set_error (error,
+		             NM_VPN_PLUGIN_ERROR,
+		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+		             "Cannot create run-dir %s (%s)",
+		             RUNDIR, g_strerror (errsv));
+		return NULL;
+	}
+
+	return g_strdup_printf (RUNDIR"/nm-openvpn-%s",
+	                        nm_connection_get_uuid (connection));
+}
+
 #define MAX_GROUPS 128
 static gboolean
 is_dir_writable (const char *dir, const char *user)
@@ -1090,20 +1110,46 @@ check_chroot_dir_usability (const char *chdir, const char *user)
 
 static gboolean
 nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
-                                 NMSettingVpn *s_vpn,
-                                 const char *default_username,
-                                 const char *uuid,
+                                 NMConnection *connection,
                                  GError **error)
 {
 	NMOpenvpnPluginPrivate *priv = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin);
-	const char *openvpn_binary, *auth, *connection_type, *tmp, *tmp2, *tmp3, *tmp4;
-	GPtrArray *args;
+	const char *openvpn_binary, *auth, *tmp, *tmp2, *tmp3, *tmp4;
+	gs_unref_ptrarray GPtrArray *args = NULL;
 	GPid pid;
 	gboolean dev_type_is_tap;
 	char *stmp;
 	const char *defport, *proto_tcp;
 	const char *nm_openvpn_user, *nm_openvpn_group, *nm_openvpn_chroot;
-	gchar *bus_name;
+	gs_free char *bus_name = NULL;
+	NMSettingVpn *s_vpn;
+	const char *connection_type;
+
+	s_vpn = nm_connection_get_setting_vpn (connection);
+	if (!s_vpn) {
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_INVALID_CONNECTION,
+		                     _("Could not process the request because the VPN connection settings were invalid."));
+		return FALSE;
+	}
+
+	connection_type = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CONNECTION_TYPE);
+	if (!validate_connection_type (connection_type)) {
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+		                     _("Invalid connection type."));
+		return FALSE;
+	}
+
+	/* Validate the properties */
+	if (!nm_openvpn_properties_validate (s_vpn, error))
+		return FALSE;
+
+	/* Validate secrets */
+	if (!nm_openvpn_secrets_validate (s_vpn, error))
+		return FALSE;
 
 	/* Find openvpn */
 	openvpn_binary = nm_find_openvpn ();
@@ -1126,16 +1172,8 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		}
 	}
 
-	connection_type = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CONNECTION_TYPE);
-	if (!validate_connection_type (connection_type)) {
-		g_set_error_literal (error,
-		                     NM_VPN_PLUGIN_ERROR,
-		                     NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-		                     _("Invalid connection type."));
-		return FALSE;
-	}
+	args = g_ptr_array_new_with_free_func (g_free);
 
-	args = g_ptr_array_new ();
 	add_openvpn_arg (args, openvpn_binary);
 
 	defport = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PORT);
@@ -1148,63 +1186,61 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE);
 	if (tmp && *tmp) {
+		gs_free char *tmp_clone = NULL;
+		char *tmp_remaining;
 		char *tok, *port, *proto;
-		char *tmp_dup = strdup (tmp);
-		while ((tok = strsep (&tmp_dup, " \t,")) != NULL) {
-			if (*tok) {
-				port = strchr (tok, ':');
-				proto = port ? strchr (port + 1, ':') : NULL;
-				if (port)
-					*port++ = '\0';
-				if (proto)
-					*proto++ = '\0';
 
-				add_openvpn_arg (args, "--remote");
-				add_openvpn_arg (args, tok);
-				if (port) {
-					if (!add_openvpn_arg_int (args, port)) {
-						g_set_error (error,
-						             NM_VPN_PLUGIN_ERROR,
-						             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-						             _("Invalid port number '%s'."), port);
-						free_openvpn_args (args);
-						g_free (tmp_dup);
-						return FALSE;
-					}
-				} else if (defport) {
-					if (!add_openvpn_arg_int (args, defport)) {
-						g_set_error (error,
-						             NM_VPN_PLUGIN_ERROR,
-						             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-						             _("Invalid port number '%s'."),
-						             defport);
-						free_openvpn_args (args);
-						return FALSE;
-					}
-				} else
-					add_openvpn_arg (args, "1194"); /* default IANA port */
+		tmp_remaining = tmp_clone = g_strdup (tmp);
+		while ((tok = strsep (&tmp_remaining, " \t,")) != NULL) {
+			if (!*tok)
+				continue;
 
-				if (proto) {
-					if (!strcmp (proto, "udp"))
-							add_openvpn_arg (args, proto);
-					else if (!strcmp (proto, "tcp"))
-							add_openvpn_arg (args, "tcp-client");
-					else {
-						g_set_error (error,
-						             NM_VPN_PLUGIN_ERROR,
-						             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-						             _("Invalid proto '%s'."), proto);
-						free_openvpn_args (args);
-						g_free (tmp_dup);
-						return FALSE;
-					}
-				} else if (proto_tcp && !strcmp (proto_tcp, "yes"))
+			port = strchr (tok, ':');
+			proto = port ? strchr (port + 1, ':') : NULL;
+			if (port)
+				*port++ = '\0';
+			if (proto)
+				*proto++ = '\0';
+
+			add_openvpn_arg (args, "--remote");
+			add_openvpn_arg (args, tok);
+			if (port) {
+				if (!add_openvpn_arg_int (args, port)) {
+					g_set_error (error,
+					             NM_VPN_PLUGIN_ERROR,
+					             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+					             _("Invalid port number '%s'."), port);
+					return FALSE;
+				}
+			} else if (defport) {
+				if (!add_openvpn_arg_int (args, defport)) {
+					g_set_error (error,
+					             NM_VPN_PLUGIN_ERROR,
+					             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+					             _("Invalid port number '%s'."),
+					             defport);
+					return FALSE;
+				}
+			} else
+				add_openvpn_arg (args, "1194"); /* default IANA port */
+
+			if (proto) {
+				if (!strcmp (proto, "udp"))
+					add_openvpn_arg (args, proto);
+				else if (!strcmp (proto, "tcp"))
 					add_openvpn_arg (args, "tcp-client");
-				else
-					add_openvpn_arg (args, "udp");
-			}
+				else {
+					g_set_error (error,
+					             NM_VPN_PLUGIN_ERROR,
+					             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+					             _("Invalid proto '%s'."), proto);
+					return FALSE;
+				}
+			} else if (proto_tcp && !strcmp (proto_tcp, "yes"))
+				add_openvpn_arg (args, "tcp-client");
+			else
+				add_openvpn_arg (args, "udp");
 		}
-		g_free (tmp_dup);
 	}
 
 	/* Remote random */
@@ -1241,14 +1277,39 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 				         NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 				         _("Invalid proxy type '%s'."),
 				         tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
 
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMP_LZO);
-	if (tmp && !strcmp (tmp, "yes"))
+
+	/* openvpn understands 4 different modes for --comp-lzo, which have
+	 * different meaning:
+	 *  1) no --comp-lzo option
+	 *  2) --comp-lzo yes
+	 *  3) --comp-lzo [adaptive]
+	 *  4) --comp-lzo no
+	 *
+	 * In the past, nm-openvpn only supported 1) and 2) by having no
+	 * comp-lzo connection setting or "comp-lzo=yes", respectively.
+	 *
+	 * However, old plasma-nm would set "comp-lzo=no" in the connection
+	 * to mean 1). Thus, "comp-lzo=no" is spoiled to mean 4) in order
+	 * to preserve backward compatibily.
+	 * We use instead a special value "no-by-default" to express "no".
+	 *
+	 * See bgo#769177
+	 */
+	if (NM_IN_STRSET (tmp, "no")) {
+		/* means no --comp-lzo option. */
+		tmp = NULL;
+	} else if (NM_IN_STRSET (tmp, "no-by-default"))
+		tmp = "no";
+
+	if (NM_IN_STRSET (tmp, "yes", "no", "adaptive")) {
 		add_openvpn_arg (args, "--comp-lzo");
+		add_openvpn_arg (args, tmp);
+	}
 
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_FLOAT);
 	if (tmp && !strcmp (tmp, "yes"))
@@ -1264,7 +1325,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Invalid ping duration '%s'."),
 			             tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
@@ -1278,7 +1338,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Invalid ping-exit duration '%s'."),
 			             tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
@@ -1292,13 +1351,29 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Invalid ping-restart duration '%s'."),
 			             tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
 
 	add_openvpn_arg (args, "--nobind");
 
+	/* max routes allowed from openvpn server */
+	tmp = nm_setting_vpn_get_data_item(s_vpn, NM_OPENVPN_KEY_MAX_ROUTES);
+	if (tmp) {
+		/* max-routes option is deprecated in 2.4 release
+		 * https://github.com/OpenVPN/openvpn/commit/d0085293e709c8a722356cfa68ad74c962aef9a2
+		 */
+		add_openvpn_arg (args, "--max-routes");
+		if (!add_openvpn_arg_int (args, tmp)) {
+			g_set_error (error,
+			             NM_VPN_PLUGIN_ERROR,
+			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+			             _("Invalid max-routes argument '%s'."),
+			             tmp);
+			return FALSE;
+		}
+	}
+
 	/* Device and device type, defaults to tun */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV);
 	tmp2 = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV_TYPE);
@@ -1334,6 +1409,12 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		add_openvpn_arg (args, tmp);
 	}
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_CIPHER);
+	if (tmp && tmp[0]) {
+		add_openvpn_arg (args, "--tls-cipher");
+		add_openvpn_arg (args, tmp);
+	}
+
 	/* Keysize */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEYSIZE);
 	if (tmp && strlen (tmp)) {
@@ -1344,7 +1425,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Invalid keysize '%s'."),
 			             tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
@@ -1370,22 +1450,47 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	/* tls-remote */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_REMOTE);
 	if (tmp && strlen (tmp)) {
-                add_openvpn_arg (args, "--tls-remote");
-                add_openvpn_arg (args, tmp);
+		add_openvpn_arg (args, "--tls-remote");
+		add_openvpn_arg (args, tmp);
+	}
+
+	/* verify-x509-name */
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_VERIFY_X509_NAME);
+	if (tmp && strlen (tmp)) {
+		const char *name;
+		gs_free char *type = NULL;
+
+		name = strchr (tmp, ':');
+		if (name) {
+			type = g_strndup (tmp, name - tmp);
+			name++;
+		} else
+			name = tmp;
+
+		if (!name[0] || !g_utf8_validate(name, -1, NULL)) {
+			g_set_error (error, NM_VPN_PLUGIN_ERROR,
+			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+			             _("Invalid verify-x509-name."));
+			return FALSE;
+		}
+
+		add_openvpn_arg (args, "--verify-x509-name");
+		add_openvpn_arg (args, name);
+		add_openvpn_arg (args, type ?: "subject");
 	}
 
 	/* remote-cert-tls */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_CERT_TLS);
 	if (tmp && strlen (tmp)) {
-                add_openvpn_arg (args, "--remote-cert-tls");
-                add_openvpn_arg (args, tmp);
+		add_openvpn_arg (args, "--remote-cert-tls");
+		add_openvpn_arg (args, tmp);
 	}
 
 	/* ns-cert-type */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_NS_CERT_TYPE);
 	if (tmp && tmp[0]) {
-                add_openvpn_arg (args, "--ns-cert-type");
-                add_openvpn_arg (args, tmp);
+		add_openvpn_arg (args, "--ns-cert-type");
+		add_openvpn_arg (args, tmp);
 	}
 
 	/* Reneg seconds */
@@ -1401,7 +1506,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Invalid reneg seconds '%s'."),
 			             tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	} else {
@@ -1414,11 +1518,14 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		add_openvpn_arg (args, "0");
 	}
 
-	if (gl.debug) {
+	if (gl.log_level_ovpn >= 0) {
+		char buf[20];
+
 		add_openvpn_arg (args, "--verb");
-		add_openvpn_arg (args, "10");
-	} else {
-		/* Syslog */
+		add_openvpn_arg (args, nm_sprintf_buf (buf, "%d", gl.log_level_ovpn));
+	}
+
+	if (gl.log_syslog) {
 		add_openvpn_arg (args, "--syslog");
 		add_openvpn_arg (args, "nm-openvpn");
 	}
@@ -1433,7 +1540,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Invalid TUN MTU size '%s'."),
 			             tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
@@ -1448,7 +1554,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Invalid fragment size '%s'."),
 			             tmp);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
@@ -1469,8 +1574,11 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	/* Up script, called when connection has been established or has been restarted */
 	add_openvpn_arg (args, "--up");
 	g_object_get (plugin, NM_VPN_SERVICE_PLUGIN_DBUS_SERVICE_NAME, &bus_name, NULL);
-	stmp = g_strdup_printf ("%s%s --bus-name %s %s --", NM_OPENVPN_HELPER_PATH, gl.debug ? " --helper-debug" : "",
-	                        bus_name, dev_type_is_tap ? "--tap" : "--tun");
+	stmp = g_strdup_printf ("%s --debug %d %ld --bus-name %s %s --",
+	                        NM_OPENVPN_HELPER_PATH,
+	                        gl.log_level, (long) getpid(),
+	                        bus_name,
+	                        dev_type_is_tap ? "--tap" : "--tun");
 	add_openvpn_arg (args, stmp);
 	g_free (stmp);
 	add_openvpn_arg (args, "--up-restart");
@@ -1480,9 +1588,11 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	add_openvpn_arg (args, "--persist-tun");
 
 	/* Management socket for localhost access to supply username and password */
+	g_clear_pointer (&priv->mgt_path, g_free);
+	priv->mgt_path = mgt_path_create (connection, error);
+	if (!priv->mgt_path)
+		return FALSE;
 	add_openvpn_arg (args, "--management");
-	g_warn_if_fail (priv->mgt_path == NULL);
-	priv->mgt_path = g_strdup_printf (LOCALSTATEDIR "/run/NetworkManager/nm-openvpn-%s", uuid);
 	add_openvpn_arg (args, priv->mgt_path);
 	add_openvpn_arg (args, "unix");
 	add_openvpn_arg (args, "--management-client-user");
@@ -1523,7 +1633,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			                     NM_VPN_PLUGIN_ERROR,
 			                     NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			                     _("Missing required local IP address for static key mode."));
-			free_openvpn_args (args);
 			return FALSE;
 		}
 		add_openvpn_arg (args, tmp);
@@ -1535,7 +1644,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			                     NM_VPN_PLUGIN_ERROR,
 			                     NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			                     _("Missing required remote IP address for static key mode."));
-			free_openvpn_args (args);
 			return FALSE;
 		}
 		add_openvpn_arg (args, tmp);
@@ -1561,21 +1669,15 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 		             _("Unknown connection type '%s'."),
 		             connection_type);
-		free_openvpn_args (args);
 		return FALSE;
 	}
 
-	/* Allow openvpn to be run as a specified user:group and drop privileges. */
-	nm_openvpn_user = getenv ("NM_OPENVPN_USER");
-	nm_openvpn_group = getenv ("NM_OPENVPN_GROUP");
-	nm_openvpn_chroot = getenv ("NM_OPENVPN_CHROOT");
-	if (!nm_openvpn_user)
-		nm_openvpn_user = NM_OPENVPN_USER;
-	if (!nm_openvpn_group)
-		nm_openvpn_group = NM_OPENVPN_GROUP;
-	if (!nm_openvpn_chroot)
-		nm_openvpn_chroot = NM_OPENVPN_CHROOT;
-
+	/* Allow openvpn to be run as a specified user:group.
+	 *
+	 * We do this by default. The only way to disable it is by setting
+	 * empty environment variables NM_OPENVPN_USER and NM_OPENVPN_GROUP. */
+	nm_openvpn_user = getenv ("NM_OPENVPN_USER") ?: NM_OPENVPN_USER;
+	nm_openvpn_group = getenv ("NM_OPENVPN_GROUP") ?: NM_OPENVPN_GROUP;
 	if (*nm_openvpn_user) {
 		if (getpwnam (nm_openvpn_user)) {
 			add_openvpn_arg (args, "--user");
@@ -1586,7 +1688,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("User '%s' not found, check NM_OPENVPN_USER."),
 			             nm_openvpn_user);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
@@ -1600,16 +1701,19 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 			             _("Group '%s' not found, check NM_OPENVPN_GROUP."),
 			             nm_openvpn_group);
-			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
+
+	/* we try to chroot be default. The only way to disable that is by
+	 * setting the an empty environment variable NM_OPENVPN_CHROOT. */
+	nm_openvpn_chroot = getenv ("NM_OPENVPN_CHROOT") ?: NM_OPENVPN_CHROOT;
 	if (*nm_openvpn_chroot) {
 		if (check_chroot_dir_usability (nm_openvpn_chroot, nm_openvpn_user)) {
 			add_openvpn_arg (args, "--chroot");
 			add_openvpn_arg (args, nm_openvpn_chroot);
 		} else
-			_LOG_W ("Directory '%s' not usable for chroot by '%s', openvpn will not be chrooted.",
+			_LOGW ("Directory '%s' not usable for chroot by '%s', openvpn will not be chrooted.",
 			        nm_openvpn_chroot, nm_openvpn_user);
 	}
 
@@ -1618,15 +1722,12 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	{
 		gs_free char *cmd = NULL;
 
-		_LOG_d ("EXEC: '%s'", (cmd = g_strjoinv (" ", (char **) args->pdata)));
+		_LOGD ("EXEC: '%s'", (cmd = g_strjoinv (" ", (char **) args->pdata)));
 	}
 
 	if (!g_spawn_async (NULL, (char **) args->pdata, NULL,
-	                    G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, error)) {
-		free_openvpn_args (args);
+	                    G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, error))
 		return FALSE;
-	}
-	free_openvpn_args (args);
 
 	pids_pending_add (pid, plugin);
 
@@ -1644,7 +1745,8 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	    || nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME)) {
 
 		priv->io_data = g_malloc0 (sizeof (NMOpenvpnPluginIOData));
-		update_io_data_from_vpn_setting (priv->io_data, s_vpn, default_username);
+		update_io_data_from_vpn_setting (priv->io_data, s_vpn,
+		                                 nm_setting_vpn_get_user_name (s_vpn));
 		nm_openvpn_schedule_connect_timer (plugin);
 	}
 
@@ -1715,7 +1817,12 @@ real_disconnect (NMVpnServicePlugin *plugin,
 {
 	NMOpenvpnPluginPrivate *priv = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin);
 
-	g_clear_pointer (&priv->mgt_path, g_free);
+	if (priv->mgt_path) {
+		/* openvpn does not cleanup the management socket upon exit,
+		 * possibly it could not even because it changed user */
+		(void) unlink (priv->mgt_path);
+		g_clear_pointer (&priv->mgt_path, g_free);
+	}
 
 	if (priv->pid) {
 		pids_pending_send_sigterm (priv->pid);
@@ -1731,51 +1838,16 @@ _connect_common (NMVpnServicePlugin   *plugin,
                  GVariant      *details,
                  GError       **error)
 {
-	NMSettingVpn *s_vpn;
-	const char *connection_type;
-	const char *user_name;
+	GError *local = NULL;
 
-	if (!real_disconnect (plugin, error)) {
-		_LOG_W ("Could not clean up previous daemon run: %s", (*error)->message);
-		g_clear_error (error);
+	if (!real_disconnect (plugin, &local)) {
+		_LOGW ("Could not clean up previous daemon run: %s", local->message);
+		g_error_free (local);
 	}
 
-	s_vpn = nm_connection_get_setting_vpn (connection);
-	if (!s_vpn) {
-		g_set_error_literal (error,
-		                     NM_VPN_PLUGIN_ERROR,
-		                     NM_VPN_PLUGIN_ERROR_INVALID_CONNECTION,
-		                     _("Could not process the request because the VPN connection settings were invalid."));
-		return FALSE;
-	}
-
-	connection_type = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CONNECTION_TYPE);
-	if (!validate_connection_type (connection_type)) {
-		g_set_error_literal (error,
-		                     NM_VPN_PLUGIN_ERROR,
-		                     NM_VPN_PLUGIN_ERROR_INVALID_CONNECTION,
-		                     _("Could not process the request because the openvpn connection type was invalid."));
-		return FALSE;
-	}
-
-	/* Validate the properties */
-	if (!nm_openvpn_properties_validate (s_vpn, error))
-		return FALSE;
-
-	/* Validate secrets */
-	if (!nm_openvpn_secrets_validate (s_vpn, error))
-		return FALSE;
-
-	/* Finally try to start OpenVPN */
-	user_name = nm_setting_vpn_get_user_name (s_vpn);
-	if (!nm_openvpn_start_openvpn_binary (NM_OPENVPN_PLUGIN (plugin),
-	                                      s_vpn,
-	                                      user_name,
-	                                      nm_connection_get_uuid (connection),
-	                                      error))
-		return FALSE;
-
-	return TRUE;
+	return nm_openvpn_start_openvpn_binary (NM_OPENVPN_PLUGIN (plugin),
+	                                        connection,
+	                                        error);
 }
 
 static gboolean
@@ -1812,8 +1884,8 @@ real_need_secrets (NMVpnServicePlugin *plugin,
 	g_return_val_if_fail (NM_IS_VPN_SERVICE_PLUGIN (plugin), FALSE);
 	g_return_val_if_fail (NM_IS_CONNECTION (connection), FALSE);
 
-	if (_LOG_d_enabled ()) {
-		_LOG_d ("connection -------------------------------------");
+	if (_LOGD_enabled ()) {
+		_LOGD ("connection -------------------------------------");
 		nm_connection_dump (connection);
 	}
 
@@ -1860,7 +1932,7 @@ real_new_secrets (NMVpnServicePlugin *plugin,
 		return FALSE;
 	}
 
-	_LOG_d ("VPN received new secrets; sending to management interface");
+	_LOGD ("VPN received new secrets; sending to management interface");
 
 	update_io_data_from_vpn_setting (priv->io_data, s_vpn, NULL);
 
@@ -1875,7 +1947,7 @@ real_new_secrets (NMVpnServicePlugin *plugin,
 
 	/* Request new secrets if we need any */
 	if (message) {
-		_LOG_d ("Requesting new secrets: '%s'", message);
+		_LOGD ("Requesting new secrets: '%s'", message);
 		nm_vpn_service_plugin_secrets_required (plugin, message, (const char **) hints);
 	}
 	if (hints)
@@ -1951,12 +2023,13 @@ nm_openvpn_plugin_new (const char *bus_name)
 
 	plugin =  (NMOpenvpnPlugin *) g_initable_new (NM_TYPE_OPENVPN_PLUGIN, NULL, &error,
 	                                              NM_VPN_SERVICE_PLUGIN_DBUS_SERVICE_NAME, bus_name,
+	                                              NM_VPN_SERVICE_PLUGIN_DBUS_WATCH_PEER, !gl.debug,
 	                                              NULL);
 
 	if (plugin) {
 		g_signal_connect (G_OBJECT (plugin), "state-changed", G_CALLBACK (plugin_state_changed), NULL);
 	} else {
-		_LOG_W ("Failed to initialize a plugin instance: %s", error->message);
+		_LOGW ("Failed to initialize a plugin instance: %s", error->message);
 		g_error_free (error);
 	}
 
@@ -2019,14 +2092,39 @@ main (int argc, char *argv[])
 	                                "OpenVPN capability to NetworkManager."));
 
 	if (!g_option_context_parse (opt_ctx, &argc, &argv, &error)) {
-		_LOG_W ("Error parsing the command line options: %s", error->message);
+		g_printerr ("Error parsing the command line options: %s\n", error->message);
 		g_option_context_free (opt_ctx);
 		g_clear_error (&error);
 		exit (1);
 	}
 	g_option_context_free (opt_ctx);
 
-	_LOG_d ("nm-openvpn-service (version " DIST_VERSION ") starting...");
+	gl.log_level = _nm_utils_ascii_str_to_int64 (getenv ("NM_VPN_LOG_LEVEL"),
+	                                             10, 0, LOG_DEBUG, -1);
+	if (gl.log_level >= 0) {
+		if (gl.log_level >= LOG_DEBUG)
+			gl.log_level_ovpn = 10;
+		else if (gl.log_level >= LOG_INFO)
+			gl.log_level_ovpn = 5;
+		else if (gl.log_level > 0)
+			gl.log_level_ovpn = 2;
+		else
+			gl.log_level_ovpn = 1;
+	} else if (gl.debug)
+		gl.log_level_ovpn = 10;
+	else {
+		/* the default level is already "--verb 1", which is fine for us. */
+		gl.log_level_ovpn = -1;
+	}
+
+	if (gl.log_level < 0)
+		gl.log_level = gl.debug ? LOG_INFO : LOG_NOTICE;
+
+	gl.log_syslog = _nm_utils_ascii_str_to_int64 (getenv ("NM_VPN_LOG_SYSLOG"),
+	                                              10, 0, 1,
+	                                              gl.debug ? 0 : 1);
+
+	_LOGD ("nm-openvpn-service (version " DIST_VERSION ") starting...");
 
 	if (   !g_file_test ("/sys/class/misc/tun", G_FILE_TEST_EXISTS)
 	    && (system ("/sbin/modprobe tun") == -1))
